/*
 * Copyright (c) 2017, 2021, Oracle and/or its affiliates.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2.0, as
 * published by the Free Software Foundation.
 *
 * This program is also distributed with certain software (including
 * but not limited to OpenSSL) that is licensed under separate terms,
 * as designated in a particular file or component or in included license
 * documentation.  The authors of MySQL hereby grant you an
 * additional permission to link the program and your derivative works
 * with the separately licensed software that they have included with
 * MySQL.
 *
 * Without limiting anything contained in the foregoing, this file,
 * which is part of MySQL Connector/Node.js, is also subject to the
 * Universal FOSS Exception, version 1.0, a copy of which can be found at
 * http://oss.oracle.com/licenses/universal-foss-exception.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License, version 2.0, for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
 */

'use strict';

const parseQueryParameters = require('./parseQueryParameters');

module.exports = parse;

function parseItemList (input) {
    if (input === '[]') {
        return [];
    }

    const match = (input || '[]').match(/^\[([^\]]*)\]$/);
    const list = !match ? [] : match[1];

    if (!list.length) {
        return undefined;
    }

    return list.split(',');
}

/**
 * Parse SSL/TLS options.
 * @private
 * @param {string} input - URI querystring
 * @returns {Object} Security section dictionary for URI properties object.
 */
function parse (input) {
    // TODO(Rui): use default agument values on node >= 6.0.0
    const match = (input || '').trim().match(/^\?([^#]+)/) || [];
    const params = parseQueryParameters(match[1], { allowDuplicates: false, ignoreCase: ['ssl-mode'] });

    const isSecure = params['ssl-mode'] !== 'disabled';
    const options = Object.assign({}, params, { 'ssl-enabled': isSecure });

    const versions = parseItemList(params['tls-versions']);
    const ciphersuites = parseItemList(params['tls-ciphersuites']);

    if (versions) {
        options['tls-versions'] = versions;
    }

    if (ciphersuites) {
        options['tls-ciphersuites'] = ciphersuites;
    }

    delete options['ssl-mode'];

    return Object.keys(options).reduce((result, key) => {
        const match = key.trim().match(/^ssl-(.+)$/) || key.trim().match(/^tls-(.+)$/) || [];

        return !match[1] ? result : Object.assign(result, { [match[1]]: options[key] });
    }, {});
}
